# Constants as Strings
ADDRESS_ZERO = "0x0000000000000000000000000000000000000000"
MAX_INT = "0xffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff"
HASH_ZERO = "0x0000000000000000000000000000000000000000000000000000000000000000"

# Constants as Int
WEI_PER_ETHER = 1000000000000000000

# Grouped constants as Tuples
DYNAMIC_FEE_TXN_PARAMS = ("maxFeePerGas", "maxPriorityFeePerGas")
